# -----------------------------------------------------------------------------
# Copyright (c) 2021- Spyder Project Contributors
#
# Released under the terms of the MIT License
# (see LICENSE.txt in the project root directory for details)
# -----------------------------------------------------------------------------

"""
The current API for Spyder plugins, originally introduced in Spyder 5.

All Spyder plugins must inherit from the classes present in this file.

.. deprecated:: 6.2

    This module will be renamed to the private :mod:`!_api` in Spyder 6.2,
    while the current public name will become an alias raising a
    :exc:`DeprecationWarning` on import and be removed in 7.0.

    Use the canonical location, the top-level :mod:`spyder.api.plugins` instead
    which already exports all the public objects of this module.

    For example, ``import spyder.api.plugins`` instead of
    ``import spyder.api.plugins.new_api``, or
    ``from spyder.api.plugins import SpyderDockablePlugin`` instead of
    ``from spyder.api.plugins.new_api import SpyderDockablePlugin``.
"""

from __future__ import annotations

# Standard library imports
import inspect
import logging
import os
import os.path as osp
import sys
from typing import TYPE_CHECKING

# Third party imports
from qtpy.QtCore import QObject, Qt, Signal, Slot
from qtpy.QtGui import QCursor, QFont, QIcon
from qtpy.QtWidgets import QApplication, QWidget

# Local imports
from spyder.api.config.mixins import BasicTypes, SpyderConfigurationObserver
from spyder.api.exceptions import SpyderAPIError
from spyder.api.fonts import SpyderFontType
from spyder.api.plugin_registration.mixins import SpyderPluginObserver
from spyder.api.widgets.main_widget import PluginMainWidget
from spyder.api.widgets.mixins import SpyderActionMixin, SpyderWidgetMixin
from spyder.app.cli_options import get_options
from spyder.config.gui import get_color_scheme, get_font
from spyder.config.user import NoDefault
from spyder.utils.icon_manager import ima
from spyder.utils.image_path_manager import IMAGE_PATH_MANAGER

# Package imports
from spyder.api.plugins.enum import Plugins

if TYPE_CHECKING:
    import argparse

    import spyder.app.mainwindow
    import spyder.config.manager
    import spyder.config.types
    import spyder.config.user
    import spyder.plugins.layout.api
    import spyder.utils.qthelpers
    import spyder.widgets.dock
    from spyder.api.preferences import PluginConfigPage, SpyderPreferencesTab
    from spyder.api.widgets.main_container import PluginMainContainer
    from spyder.api.widgets.menus import PluginMainWidgetOptionsMenu


# Logging
logger = logging.getLogger(__name__)


class SpyderPluginV2(
    QObject,
    SpyderActionMixin,
    SpyderConfigurationObserver,
    SpyderPluginObserver,
):
    """
    Base class for all Spyder plugins.

    Use this class directly for plugins that extend functionality without a
    dockable widget. To create a plugin that adds a new pane to the interface,
    use :class:`SpyderDockablePlugin`.
    """

    # --- API: Mandatory attributes ------------------------------------------
    # ------------------------------------------------------------------------

    NAME: str = None  # type: ignore # Must always be set for plugin subclasses
    """
    Name of the plugin that will be used to refer to it.

    This name must be unique and will only be loaded once.
    """

    # --- API: Optional attributes ------------------------------------------
    # -----------------------------------------------------------------------

    REQUIRES: list[str] = []
    """
    List of names of required dependencies for this plugin.

    .. note::

        Plugin names are defined in the :class:`~spyder.api.plugins.Plugins`
        pseudo-enum class.

    Examples
    --------

    .. code-block:: python

        REQUIRES = [Plugins.Plots, Plugins.IPythonConsole, ...]
    """

    OPTIONAL: list[str] = []
    """
    List of names of optional dependencies for this plugin.

    Optional dependencies are useful for when a plugin can offer specific
    features by connecting to another plugin, but does not depend on the
    other plugin for its core functionality. For example, the :guilabel:`Help`
    plugin might render information from the :guilabel:`Editor`,
    :guilabel:`IPython Console` or another source, but it does not depend on
    either of those plugins to work.

    .. note::

        Plugin names are defined in the :class:`~spyder.api.plugins.Plugins`
        pseudo-enum class.

    Examples
    --------

    .. code-block:: python

        OPTIONAL = [Plugins.Plots, Plugins.IPythonConsole, ...]
    """

    CONTAINER_CLASS: type[PluginMainContainer] | None = None
    """
    Container class object to instantiate for the plugin.

    This must subclass a
    :class:`~spyder.api.widgets.main_container.PluginMainContainer` for
    non-dockable plugins that create a widget, like a status bar widget,
    a toolbar, a menu, etc.

    For non-dockable plugins that do not define widgets of any kind,
    for example a plugin that only exposes a configuration page,
    this can be ``None``,
    """

    CONF_SECTION: str | None = None
    """
    Name of the Spyder configuration section for the plugin's data.

    Used to record the plugin's permanent data in Spyder config system
    (i.e. in :file:`spyder.ini`).
    """

    CONF_FILE: bool = True
    """
    Whether to use a separate configuration file for the plugin.

    Should always be set to ``True`` for external plugins.

    If ``True`` (the default), use a separate configuration file.
    If ``False``, use the main Spyder config file.
    """

    CONF_DEFAULTS: (
        list[
            tuple[
                str,
                dict[
                    spyder.config.types.ConfigurationKey,
                    BasicTypes | spyder.config.user.NoDefault,
                ],
            ]
        ]
        | None
    ) = None
    """
    Define configuration defaults if using a separate file.

    List of tuples, with the first item in the tuple being the section
    name and the second item being the default options dictionary.

    Examples
    --------

    .. code-block:: python

        CONF_DEFAULTS = [
            (
                "section-name",
                {
                    "option-1": "some-value",
                    "option-2": True,
                },
            ),
            (
                "another-section-name",
                {
                    "option-3": "some-other-value",
                    "option-4": [1, 2, 3],
                },
            ),
        ]
    """

    CONF_VERSION: str | None = None
    """
    Define the configuration version if using a separate file.

    * If you want to *change* the default value of a current option, you need
      to do a MINOR update in config version, e.g. from ``3.0.0`` to ``3.1.0``.
    * If you want to *remove* options that are no longer needed or if you
      want to *rename* options, then you need to do a MAJOR update in
      version, e.g. from ``3.0.0`` to ``4.0.0``.
    * You don't need to touch this value if you're just adding a new option.
    """

    CONF_WIDGET_CLASS: type[PluginConfigPage] | None = None
    """
    Widget to be used as this plugin's entry in Spyder Preferences dialog.

    If ``None``, the default, the plugin will not have a configuration
    page in Spyder's :guilabel:`Preferences`.
    """

    ADDITIONAL_CONF_OPTIONS: (
        dict[
            str,
            dict[
                spyder.config.types.ConfigurationKey,
                BasicTypes | spyder.config.user.NoDefault,
            ],
        ]
        | None
    ) = None
    """
    Configuration options added by this plugin for other plugins.

    Examples
    --------

    .. code-block:: python

        ADDITIONAL_CONF_OPTIONS = {
            "section": {
                "option_1": True,
                "option_2": "default_value",
            },
        }
    """

    ADDITIONAL_CONF_TABS: (
        dict[str, list[type[SpyderPreferencesTab]]] | None
    ) = None
    """
    Define additional tabs of options for other plugins' configuration pages.

    All configuration tabs should inherit from
    :class:`~spyder.api.preferences.SpyderPreferencesTab`.

    Examples
    --------

    .. code-block:: python

        ADDITIONAL_CONF_TABS = {"plugin_name": [MyPluginsSpyderPrefsTab]}
    """

    CUSTOM_LAYOUTS: list[
        type[spyder.plugins.layout.api.BaseGridLayoutType]
    ] = []
    """
    Define custom layout classes that the plugin wants to be registered.

    The custom classes should extend from
    :class:`spyder.plugins.layout.api.BaseGridLayoutType`.
    """

    IMG_PATH: str | None = None
    """
    Path for images relative to the plugin path.

    A Python package can include one or several Spyder plugins. If the latter,
    the package may be using images from a directory outside that of the
    plugin itself.
    """

    MONOSPACE_FONT_SIZE_DELTA: int = 0
    """
    The monospace (code) font size delta relative to Spyder default, in points.
    """

    INTERFACE_FONT_SIZE_DELTA: int = 0
    """
    The interface (UI) font size delta relative to Spyder default, in points.
    """

    CONTEXT_NAME: str | None = None
    """
    The name under which to store actions, toolbars, toolbuttons and menus.
    """

    CAN_BE_DISABLED: bool = True
    """
    Define if a plugin can be disabled in the Spyder :guilabel:`Preferences`.

    If ``False``, the plugin is considered "core" and cannot be disabled.
    ``True`` by default, meaning the plugin can be disabled.
    """

    REQUIRE_WEB_WIDGETS: bool = False
    """
    Declare whether the plugin needs to use Qt web widgets (QtWebEngine).

    Qt Web Widgets is a heavy dependency for many packagers, e.g. Conda-Forge.
    We thus ask plugins to declare whether or not they need
    web widgets to make it easier to distribute Spyder without them.

    See `Spyder issue #22196 <https://github.com/spyder-ide/spyder/pull/22196#issuecomment-2189377043>`__
    for more information.
    """

    # --- API: Signals -------------------------------------------------------
    # ------------------------------------------------------------------------
    # Signals here are automatically connected by the Spyder main window and
    # connected to the the respective global actions defined on it.
    sig_free_memory_requested: Signal = Signal()
    """
    Request the main application garbage collect deleted objects.
    """

    sig_plugin_ready: Signal = Signal()
    """
    Emitted once the plugin is initialized.
    """

    sig_quit_requested: Signal = Signal()
    """
    Request that the main Spyder application quit.
    """

    sig_restart_requested: Signal = Signal()
    """
    Request that the main Spyder application perform a restart.
    """

    sig_status_message_requested: Signal = Signal(str, int)
    """
    Request that the main application display a status bar message.

    Parameters
    ----------
    message: str
        The message to display.
    timeout: int
        The timeout before the message disappears, in milliseconds.
    """

    sig_redirect_stdio_requested: Signal = Signal(bool)
    """
    Request the main app redirect standard out/error within file pickers.
    
    This will redirect :data:`~sys.stdin`, :data:`~sys.stdout`, and
    :data:`~sys.stderr` when using :guilabel:`Open`, :guilabel:`Save`,
    and :guilabel:`Browse` dialogs within a plugin's widgets.

    Parameters
    ----------
    enable: bool
        Enable (``True``) or disable (``False``) standard input/output
        redirection.
    """

    sig_exception_occurred: Signal = Signal(dict)
    """
    Report an exception from a plugin.

    Parameters
    ----------
    error_data: dict[str, str | bool]
        The dictionary containing error data. The expected keys are:

        .. code-block:: python

            error_data = {
                "text": str,
                "is_traceback": bool,
                "repo": str,
                "title": str,
                "label": str,
                "steps": str,
            }

        The ``is_traceback`` key indicates if ``text`` contains plain text or a
        Python error traceback.

        The ``title`` and ``repo`` keys indicate how the error data should
        customize the report dialog and GitHub error submission.

        The ``label`` and ``steps`` keys allow customizing the content of the
        error dialog.
    """

    sig_mainwindow_resized: Signal = Signal("QResizeEvent")
    """
    Emitted when the main window is resized.

    Used by plugins to track main window size changes.

    Parameters
    ----------
    resize_event: QResizeEvent
        The event triggered on main window resize.
    """

    sig_mainwindow_moved: Signal = Signal("QMoveEvent")
    """
    Emitted when the main window is moved.

    Used by plugins to track main window position changes.

    Parameters
    ----------
    move_event: QMoveEvent
        The event triggered on main window move.
    """

    sig_unmaximize_plugin_requested: Signal = Signal((), (object,))
    """
    Request the main window unmaximize the currently maximized plugin, if any.

    Parameters
    ----------
    plugin_instance: SpyderDockablePlugin
        Unmaximize current plugin only if it is not ``plugin_instance``.
    """

    sig_mainwindow_state_changed: Signal = Signal(object)
    """
    Emitted when the main window state has changed (e.g. maximized/minimized).

    Parameters
    ----------
    window_state: Qt.WindowStates
        The new main window state.
    """

    sig_focused_plugin_changed: Signal = Signal(object)
    """
    Emitted when the plugin with keyboard focus changes.

    Parameters
    ----------
    plugin: SpyderDockablePlugin | None
        The plugin that currently has keyboard focus, or ``None`` if no
        dockable plugin has focus.
    """

    # ---- Private attributes
    # -------------------------------------------------------------------------

    _CONF_NAME_MAP = None
    """
    Define config name map for plugin to split config among several files.

    See :mod:`spyder.config.main`.
    """

    # ---- Private methods
    # -------------------------------------------------------------------------

    def __init__(
        self,
        parent: spyder.app.mainwindow.MainWindow,
        configuration: (
            spyder.config.manager.ConfigurationManager | None
        ) = None,
    ):
        """
        Initialize the plugin object (called automatically by Spyder).

        .. important::

            Plugins are initialized automatically by Spyder, so they shouldn't
            call or override this method directly.

        Parameters
        ----------
        parent : spyder.app.mainwindow.MainWindow
            Parent Spyder window of this plugin.
        configuration : spyder.config.manager.ConfigurationManager
            The Spyder configuration manager object for the plugin to access.

        Returns
        -------
        None
        """
        super().__init__(parent)

        # This is required since the MRO of this class does not call
        # SpyderPluginObserver and SpyderConfigurationObserver when using
        # super(), see https://fuhm.net/super-harmful/
        SpyderPluginObserver.__init__(self)
        SpyderConfigurationObserver.__init__(self)

        self._main = parent
        self._conf = configuration
        self._plugin_path = os.path.dirname(inspect.getfile(self.__class__))

        self._widget: PluginMainWidget | PluginMainContainer | None = None
        self._container: PluginMainWidget | PluginMainContainer | None = None

        self.PLUGIN_NAME: str = self.NAME
        """
        Plugin name in the action, toolbar, toolbutton & menu registries.

        Usually the same as :attr:`NAME`, but may be different from
        :attr:`CONTEXT_NAME`.
        """

        self.is_compatible: bool | None = None
        """Whether the plugin has passed Spyder's compatibility checks.

        ``True`` if it has, ``False`` if it hasn't, and ``None`` if the checks
        haven't been run yet.
        """

        self.is_registered: bool | None = None
        """Whether the plugin is enabled and registered with Spyder.

        ``True`` if it has been registered, ``False`` if it has been
        unregistered, and ``None`` if the plugin hasn't been set up yet.
        """

        if self.CONTAINER_CLASS is not None:
            self._container = container = self.CONTAINER_CLASS(
                name=self.NAME, plugin=self, parent=parent
            )

            if hasattr(container, "_setup"):
                container._setup()

            if isinstance(container, SpyderWidgetMixin):
                container.setup()
                container.update_actions()

            # Default signals to connect in main container or main widget.
            container.sig_free_memory_requested.connect(
                self.sig_free_memory_requested
            )
            container.sig_quit_requested.connect(self.sig_quit_requested)
            container.sig_restart_requested.connect(self.sig_restart_requested)
            container.sig_redirect_stdio_requested.connect(
                self.sig_redirect_stdio_requested
            )
            container.sig_exception_occurred.connect(
                self.sig_exception_occurred
            )
            container.sig_unmaximize_plugin_requested.connect(
                self.sig_unmaximize_plugin_requested
            )

            self.after_container_creation()

        # Load the custom images of the plugin
        if self.IMG_PATH:
            plugin_path = osp.join(self.get_path(), self.IMG_PATH)
            IMAGE_PATH_MANAGER.add_image_path(plugin_path)

    def _register(self, omit_conf: bool = False) -> None:
        """
        Setup and register the plugin and connect it to other plugins.
        """
        # Checks
        # --------------------------------------------------------------------
        if self.NAME is None:
            raise SpyderAPIError("A Spyder Plugin must define a `NAME`!")

        # Setup configuration
        # --------------------------------------------------------------------
        if self._conf is not None and not omit_conf:
            self._conf.register_plugin(self)

        # Signals
        # --------------------------------------------------------------------
        self.is_registered = True

        self.update_font()

    def _unregister(self) -> None:
        """
        Disconnect signals & clean up plugin to stop it while Spyder's running.
        """

        if self._conf is not None:
            self._conf.unregister_plugin(self)

        self._container = None
        self.is_compatible = None
        self.is_registered = False

    # ---- Convinience attributes as properties
    # -------------------------------------------------------------------------
    @property
    def main(self) -> spyder.app.mainwindow.MainWindow:
        """Spyder main window to which this plugin belongs; i.e. its parent."""
        return self._main

    @main.setter
    def main(self, value: spyder.app.mainwindow.MainWindow) -> None:
        """Spyder main window to which this plugin belongs; i.e. its parent."""
        self._main = value

    # ---- API: available methods
    # -------------------------------------------------------------------------
    def get_path(self) -> str:
        """
        Return the path on disk to the plugin's root module directory.

        Returns
        -------
        str
            The path to the directory containing the module with the plugin
            class (:class:`SpyderPluginV2` or :class:`SpyderDockablePlugin`).
        """
        return self._plugin_path

    def get_container(self) -> PluginMainWidget | PluginMainContainer | None:
        """
        Return the plugin's main container object.

        Returns
        -------
        PluginMainContainer | None
            The plugin's main container object (i.e. an instance of
            :attr:`CONTAINER_CLASS`), or ``None`` if it doesn't have one set
            (e.g. for simple plugins that don't define any widgets).
        """
        return self._container

    def get_configuration(
        self,
    ) -> spyder.config.manager.ConfigurationManager | None:
        """
        Return the Spyder configuration object.

        Returns
        -------
        spyder.config.manager.ConfigurationManager
            The Spyder configuration manager object.
        """
        return self._conf

    def get_main(self) -> spyder.app.mainwindow.MainWindow:
        """
        Return the Spyder main window.

        Returns
        -------
        spyder.app.mainwindow.MainWindow
            The Spyder main window.
        """
        return self._main

    def get_plugin(self, plugin_name, error=True) -> SpyderPluginV2 | None:
        """
        Get a plugin instance object by its name.

        Parameters
        ----------
        plugin_name: str
            Name of the plugin from which its instance will be returned.
        error: bool, optional
            If ``True`` (the default), raise an error if the plugin instance
            with the given ``plugin_name`` cannot be found, and the plugin
            is a required dependency of this one (listed in :attr`REQUIRES`).
            If ``False``, runtime errors finding ``plugin_name`` pass silently
            (unless ``plugin_name`` is not listed in :attr:`REQUIRES` or
            :attr:`OPTIONAL`, which raises an error unconditionally).

        Returns
        -------
        SpyderPluginV2 | None
            The plugin object with name ``plugin_name``, or ``None`` if it
            cannot be found and is either an optional dependency of this plugin
            (listed under :attr:`OPTIONAL`), or ``error`` is ``False``.

        Raises
        ------
        SpyderAPIError
            If ``plugin_name`` is not listed under either the :attr:`REQUIRES`
            or :attr:`OPTIONAL` attributes of this plugin, or ``plugin_name``
            is not found (if a required dependency and ``error`` is ``True``).
        """
        # Ensure that this plugin has the plugin corresponding to
        # `plugin_name` listed as required or optional.
        requires = set(self.REQUIRES or [])
        optional = set(self.OPTIONAL or [])
        full_set = requires | optional

        if plugin_name in full_set or Plugins.All in full_set:
            try:
                return self._main.get_plugin(plugin_name, error=error)
            except SpyderAPIError as e:
                if plugin_name in optional:
                    return None
                else:
                    raise e
        else:
            raise SpyderAPIError(
                'Plugin "{}" not part of REQUIRES or '
                "OPTIONAL requirements!".format(plugin_name)
            )

    def is_plugin_enabled(self, plugin_name: str) -> bool:
        """
        Determine if a given plugin is going to be loaded.

        Parameters
        ----------
        plugin_name : str
            The name of the plugin to check.

        Returns
        -------
        bool
            ``True`` if ``plugin_name`` will be loaded, else ``False``.
        """
        return self._main.is_plugin_enabled(plugin_name)

    def is_plugin_available(self, plugin_name: str) -> bool:
        """
        Determine if a given plugin is loaded and ready.

        Parameters
        ----------
        plugin_name : str
            The name of the plugin to check.

        Returns
        -------
        bool
            ``True`` if ``plugin_name`` is loaded and ready, else ``False``.
        """
        return self._main.is_plugin_available(plugin_name)

    def get_dockable_plugins(self) -> list[SpyderDockablePlugin]:
        """
        Get a list of dockable plugin instances that this plugin depends on.

        Only required plugins (listed under :attr:`REQUIRES`) that extend
        :class:`SpyderDockablePlugin` are returned.

        Returns
        -------
        list[SpyderDockablePlugin]
            List of dockable plugin object instances (those with graphical
            panes in the UI) that this plugin requires as dependencies.
        """
        requires = set(self.REQUIRES or [])
        dockable_plugins_required = []
        for name, plugin_instance in self._main.get_dockable_plugins():
            if (name in requires or Plugins.All in requires) and isinstance(
                plugin_instance, SpyderDockablePlugin
            ):
                dockable_plugins_required.append(plugin_instance)
        return dockable_plugins_required

    def get_conf(
        self,
        option: spyder.config.types.ConfigurationKey,
        default: spyder.config.user.NoDefault | BasicTypes = NoDefault,
        section: str | None = None,
        secure: bool = False,
    ) -> BasicTypes | None:
        """
        Retrieve an option's value from the Spyder configuration system.

        Parameters
        ----------
        option: spyder.config.types.ConfigurationKey
            Name/tuple path of the configuration option value to get.
        default: spyder.api.config.mixins.BasicTypes | spyder.config.user.NoDefault, optional
            Fallback value to return if the option is not found on the
            configuration system. No default value if not passed.
        section: str | None, optional
            Name of the configuration section to use, e.g. ``"shortcuts"``.
            If ``None``, the default, then the value of :attr:`CONF_SECTION`
            is used.
        secure: bool, optional
            If ``True``, the option will be retrieved from secure storage
            using the :mod:`!keyring` Python package. Otherwise, will be
            retrieved from Spyder's normal configuration (the default).

        Returns
        -------
        value: spyder.api.config.mixins.BasicTypes | None
            Value of ``option`` in the configuration ``section``, or ``None``
            if the Spyder configuration object is not available (typically only
            the case in tests).

        Raises
        ------
        SpyderAPIError
            If ``section`` is not passed and :attr:`CONF_SECTION` is not
            defined.
        configparser.NoOptionError
            If the ``section`` does not exist in Spyder's configuration.
        """
        if self._conf is None:
            return None

        section = self.CONF_SECTION if section is None else section
        if section is None:
            raise SpyderAPIError(
                "A spyder plugin must define a `CONF_SECTION` class "
                "attribute!"
            )
        return self._conf.get(section, option, default, secure=secure)

    @Slot(str, object)
    @Slot(str, object, str)
    def set_conf(
        self,
        option: spyder.config.types.ConfigurationKey,
        value: BasicTypes,
        section: str | None = None,
        recursive_notification: bool = True,
        secure: bool = False,
    ) -> None:
        """
        Set an option's value in the Spyder configuration system.

        Parameters
        ----------
        option: spyder.config.types.ConfigurationKey
            Name/tuple path of the configuration option to set.
        value: spyder.api.config.mixins.BasicTypes
            Value to set for the given configuration option.
        section: str | None, optional
            Name of the configuration section to use, e.g. ``"shortcuts"``.
            If ``None``, the default, then the value of :attr:`CONF_SECTION`
            is used.
        recursive_notification: bool, optional
            If ``True``, all objects that observe all changes on the
            configuration ``section`` as well as objects that observe
            partial tuple paths are notified. For example, if the
            option ``opt`` of section ``sec`` changes, then
            all observers for section ``sec`` are notified. Likewise,
            if the option ``("a", "b", "c")`` changes, then observers for
            ``("a", "b", "c")``, ``("a", "b")`` and ``"a"`` are all notified.
        secure: bool, optional
            If ``True``, the option will be saved in secure storage
            using the :mod:`!keyring` Python package. Otherwise, will be
            saved in Spyder's normal configuration (the default).

        Returns
        -------
        None

        Raises
        ------
        SpyderAPIError
            If ``section`` is not passed and :attr:`CONF_SECTION` is not
            defined.
        """
        if self._conf is not None:
            section = self.CONF_SECTION if section is None else section
            if section is None:
                raise SpyderAPIError(
                    "A spyder plugin must define a `CONF_SECTION` class "
                    "attribute!"
                )

            self._conf.set(
                section,
                option,
                value,
                recursive_notification=recursive_notification,
                secure=secure,
            )
            self.apply_conf({option}, False)

    def remove_conf(
        self,
        option: spyder.config.types.ConfigurationKey,
        section: str | None = None,
        secure: bool = False,
    ) -> None:
        """
        Remove an option from the Spyder configuration system.

        Parameters
        ----------
        option: spyder.config.types.ConfigurationKey
            Name/tuple path of the configuration option to remove.
        section: str | None, optional
            Name of the configuration section to use, e.g. ``"shortcuts"``.
            If ``None``, the default, then the value of :attr:`CONF_SECTION`
            is used.
        secure: bool, optional
            If ``True``, the option will be removed from secure storage
            using the :mod:`!keyring` Python package. Otherwise, will be
            removed from Spyder's normal configuration (the default).

        Returns
        -------
        None

        Raises
        ------
        SpyderAPIError
            If ``section`` is not passed and :attr:`CONF_SECTION` is not
            defined.
        """
        if self._conf is not None:
            section = self.CONF_SECTION if section is None else section
            if section is None:
                raise SpyderAPIError(
                    "A spyder plugin must define a `CONF_SECTION` class "
                    "attribute!"
                )

            self._conf.remove_option(section, option, secure=secure)
            self.apply_conf({option}, False)

    def apply_conf(
        self,
        options_set: set[spyder.config.types.ConfigurationKey],
        notify: bool = True,
    ) -> None:
        """
        Handle applying a set of options to this plugin's widget.

        Parameters
        ----------
        options_set: set[spyder.config.types.ConfigurationKey]
            The set of option names that were changed.
        notify: bool, optional
            If ``True``, the default, call :meth:`after_configuration_update`
            to perform plugin-specific additional operations after applying
            the configuration changes. If ``False``, don't call the method.

        Returns
        -------
        None
        """
        if self._conf is not None and options_set:
            if notify:
                self.after_configuration_update(list(options_set))

    def disable_conf(
        self,
        option: spyder.config.types.ConfigurationKey,
        section: str | None = None,
    ) -> None:
        """
        Disable notifications for an option in the Spyder configuration system.

        Parameters
        ----------
        option: spyder.config.types.ConfigurationKey
            Name of the option, either a string or a tuple of strings.
        section: str | None, optional
            Name of the configuration section to use, e.g. ``"shortcuts"``.
            If ``None``, the default, then the value of :attr:`CONF_SECTION`
            is used.

        Returns
        -------
        None

        Raises
        ------
        SpyderAPIError
            If ``section`` is not passed and :attr:`CONF_SECTION` is not
            defined.
        """
        if self._conf is not None:
            section = self.CONF_SECTION if section is None else section
            if section is None:
                raise SpyderAPIError(
                    "A spyder plugin must define a `CONF_SECTION` class "
                    "attribute!"
                )
            self._conf.disable_notifications(section, option)

    def restore_conf(
        self,
        option: spyder.config.types.ConfigurationKey,
        section: str | None = None,
    ) -> None:
        """
        Restore notifications for an option in the Spyder configuration system.

        Parameters
        ----------
        option: spyder.config.types.ConfigurationKey
            Name of the option, either a string or a tuple of strings.
        section: str | None, optional
            Name of the configuration section to use, e.g. ``"shortcuts"``.
            If ``None``, the default, then the value of :attr:`CONF_SECTION`
            is used.

        Returns
        -------
        None

        Raises
        ------
        SpyderAPIError
            If ``section`` is not passed and :attr:`CONF_SECTION` is not
            defined.
        """
        if self._conf is not None:
            section = self.CONF_SECTION if section is None else section
            if section is None:
                raise SpyderAPIError(
                    "A spyder plugin must define a `CONF_SECTION` class "
                    "attribute!"
                )
            self._conf.restore_notifications(section, option)

    @Slot(str)
    @Slot(str, int)
    def show_status_message(self, message: str, timeout: int = 0) -> None:
        """
        Show a message in the Spyder status bar.

        Parameters
        ----------
        message: str
            The message to display in the status bar.
        timeout: int
            The amount of time, in milliseconds, to display the message.
            If ``0``, the default, the message will be shown until a plugin
            calls :meth:`!show_status_message` again.

        Returns
        -------
        None
        """
        self.sig_status_message_requested.emit(message, timeout)

    def before_long_process(self, message: str) -> None:
        """
        Perform actions required before starting a long-running process.

        Shows a message in the main window's status bar,
        and changes the mouse pointer to a wait cursor.

        Parameters
        ----------
        message: str
            Message to show in the status bar when the long process starts.

        Returns
        -------
        None
        """
        if message:
            self.show_status_message(message)

        QApplication.setOverrideCursor(QCursor(Qt.WaitCursor))
        QApplication.processEvents()

    def after_long_process(self, message: str = "") -> None:
        """
        Perform actions required after starting a long-running process.

        Clears the message in the Spyder main window's status bar
        and restores the mouse pointer to the OS default.

        Parameters
        ----------
        message: str, optional
            Message to show in the status bar when the long process finishes.
            An empty message by default, clearing the previous message.

        Returns
        -------
        None
        """
        QApplication.restoreOverrideCursor()
        self.show_status_message(message, timeout=2000)
        QApplication.processEvents()

    def get_color_scheme(
        self,
    ) -> dict[str, str | tuple[str, bool, bool]] | None:
        """
        Get the Editor's current color scheme.

        This is useful to set the color scheme of all instances of
        CodeEditor used by the plugin.

        Returns
        -------
        dict[str, str | tuple[str, bool, bool]] | None
            Dictionary with properties and colors of the color scheme
            used in the Editor, or ``None`` if the Spyder configuration
            object is not available (usually only ever the case in tests).
        """
        if self._conf is None:
            return None

        return get_color_scheme(self._conf.get("appearance", "selected"))

    def initialize(self) -> None:
        """
        Initialize a plugin instance.

        .. caution::

            This method should be called to initialize the plugin, but it
            should not be overridden, since it calls :meth:`on_initialize`
            and emits the :attr:`sig_plugin_ready` signal.

        Returns
        -------
        None
        """
        self.on_initialize()
        self.sig_plugin_ready.emit()

    @staticmethod
    def create_icon(name) -> QIcon:
        """
        Retrieve an icon from the theme and icon manager.

        Parameters
        ----------
        name: str
            The name of the icon to retrieve.

        Returns
        -------
        QIcon
            The specified icon, as a :class:`QIcon` instance.
        """
        return ima.icon(name)

    @classmethod
    def get_font(cls, font_type: str) -> QFont:
        """
        Return the font object for one of the font types used in Spyder.

        All plugins in Spyder use the same, global fonts. In case a plugin
        wants to use a different font size than on the default, it can set
        the :attr:`MONOSPACE_FONT_SIZE_DELTA` or
        :attr:`INTERFACE_FONT_SIZE_DELTA` class constants.

        Parameters
        ----------
        font_type: str
            The kind of font to return the object for, as listed under the
            :class:`~spyder.api.fonts.SpyderFontType` pseudo-enum class.
            See that class' documentation for more details.

        Returns
        -------
        QFont
            :class:`QFont` object for the specified ``font_type``, to be
            passed to other Qt widgets.
        """
        if font_type == SpyderFontType.Monospace:
            font_size_delta = cls.MONOSPACE_FONT_SIZE_DELTA
        elif font_type in [
            SpyderFontType.Interface,
            SpyderFontType.MonospaceInterface,
        ]:
            font_size_delta = cls.INTERFACE_FONT_SIZE_DELTA
        else:
            raise SpyderAPIError("Unrecognized font type")

        return get_font(option=font_type, font_size_delta=font_size_delta)

    def get_command_line_options(self) -> argparse.Namespace:
        """
        Get the command line options passed by the user when starting Spyder.

        See :mod:`spyder.app.cli_options` for the option names.

        Returns
        -------
        argparse.Namespace
            Namespace of the passed option keys and values.
        """
        if self._main is not None:
            return self._main._cli_options
        else:
            # This is necessary when the plugin has no parent.
            sys_argv = [sys.argv[0]]  # Avoid options passed to pytest
            return get_options(sys_argv)[0]

    # ---- API: Mandatory methods to define
    # -------------------------------------------------------------------------
    @staticmethod
    def get_name() -> str:
        """
        Return the plugin's localized name.

        .. note::

            This method needs to be decorated with :func:`staticmethod`.

        Returns
        -------
        str
            Localized name of the plugin.
        """
        raise NotImplementedError("A plugin name must be defined!")

    @staticmethod
    def get_description() -> str:
        """
        Return the plugin's localized description.

        .. note::

            This method needs to be decorated with :func:`staticmethod`.

        Returns
        -------
        str
            Localized description of the plugin.
        """
        raise NotImplementedError("A plugin description must be defined!")

    @classmethod
    def get_icon(cls) -> QIcon:
        """
        Return the plugin's associated icon.

        .. note::

            This method needs to be decorated with :func:`classmethod` or
            :func:`staticmethod`.

        Returns
        -------
        QIcon
            The plugin's icon, as a :class:`QIcon` instance.
        """
        raise NotImplementedError("A plugin icon must be defined!")

    def on_initialize(self) -> None:
        """
        Set up the plugin.

        .. caution::

            Any calls performed in this method should not call other plugins.

        Returns
        -------
        None
        """
        raise NotImplementedError(
            f"The plugin {type(self)} is missing an implementation of "
            "on_initialize"
        )

    # ---- API: Optional methods to override
    # -------------------------------------------------------------------------
    @staticmethod
    def check_compatibility() -> tuple[bool, str]:
        """
        Check compatibility of a plugin with the user's current environment.

        Intended for plugin-specific checks, so needs to be reimplemented by
        the plugin subclass to do anything meaningful.

        Returns
        -------
        valid, message: tuple[bool, str]
            The first value, ``valid``, tells Spyder if the plugin has passed
            the compatibility test defined in this method.

            The second value, ``message``, must (if ``valid`` is ``False``)
            explain to users why the plugin was found to be incompatible
            (e.g. ``"This plugin does not work with PyQt4"``).
            It will be shown at startup in a :class:`QMessageBox`.
        """
        valid = True
        message = ""  # Note: Remember to use _('') to localize the string
        return valid, message

    def on_first_registration(self) -> None:
        """
        Actions to be performed the first time the plugin is started.

        It can also be used to perform actions that are needed only the
        first time this is loaded after installation.

        This method is called after the main window is visible.

        Returns
        -------
        None
        """
        pass

    def before_mainwindow_visible(self) -> None:
        """
        Actions to be performed after setup but before showing the main window.

        Returns
        -------
        None
        """
        pass

    def on_mainwindow_visible(self) -> None:
        """
        Actions to be performed after the main window has been shown.

        Returns
        -------
        None
        """
        pass

    def on_close(self, cancelable: bool = False) -> None:
        """
        Perform actions before the plugin is closed.

        .. caution::

            This method **must** only operate on local attributes and not
            call other plugins.

        Parameters
        ----------
        cancelable: bool, optional
            ``True`` if the close operation can potentially be canceled;
            ``False`` by default.

        Returns
        -------
        None
        """
        pass

    def can_close(self) -> bool:
        """
        Determine if the plugin can be closed.

        Returns
        -------
        close: bool
            ``True`` if the plugin can be closed, ``False`` otherwise.
        """
        return True

    def update_font(self) -> None:
        """
        Modify the font used in the plugin's interface.
        
        This must be reimplemented by plugins that need to adjust their fonts.

        The following plugins illustrate the usage of this method:
          * :mod:`spyder.plugins.help.plugin`
          * :mod:`spyder.plugins.onlinehelp.plugin`

        Returns
        -------
        None
        """
        pass

    def update_style(self) -> None:
        """
        Modify the interface styling used by the plugin.
        
        This must be reimplemented by plugins that need to adjust their style.

        Changing from the dark to the light interface theme might
        require specific styles or stylesheets to be applied. When
        the theme is changed by the user through our :guilabel:`Preferences`,
        this method will be called for all plugins.

        Returns
        -------
        None
        """
        pass

    def after_container_creation(self) -> None:
        """
        Perform necessary operations before setting up the container.

        This must be reimplemented by plugins whose containers emit signals
        that need to be connected before applying options to Spyder's config
        system.

        Returns
        -------
        None
        """
        pass

    def after_configuration_update(
        self, options: list[spyder.config.types.ConfigurationKey]
    ) -> None:
        """
        Perform additional operations after updating the plugin config options.

        This can be implemented by plugins that do not have a container and
        need to act on configuration updates.

        Parameters
        ----------
        options: list[spyder.config.types.ConfigurationKey]
            A list that contains the option names that were updated.

        Returns
        -------
        None
        """
        pass


class SpyderDockablePlugin(SpyderPluginV2):
    """
    Subclass for plugins with a dockable widget (pane) in the interface.
    """

    # ---- API: Mandatory attributes
    # -------------------------------------------------------------------------

    WIDGET_CLASS: type[PluginMainWidget] = None
    """
    This is the main widget of the dockable plugin.

    It must be a subclass of
    :class:`spyder.api.widgets.main_widget.PluginMainWidget`.
    """

    # ---- API: Optional attributes
    # -------------------------------------------------------------------------

    TABIFY: list[str] = []
    """
    Define a list of plugins next to which we want to tabify this plugin.

    Examples
    --------

    .. code-block:: python

        TABIFY = [Plugins.Editor]
    """

    DISABLE_ACTIONS_WHEN_HIDDEN: bool = True
    """
    Disable the plugin's actions in the main menu when the plugin is hidden.

    If ``True``, disable the plugin's actions when it is not shown (default);
    if ``False``, keep them enabled.
    """

    RAISE_AND_FOCUS: bool = False
    """
    Give the plugin focus on switch to plugin calls.

    If ``True``, the plugin will be given focus when switched to.
    If ``False``, the default, the plugin's widget will still be raised to
    the foreground, but it will not be given focus until the switch action
    is called a second time.
    """

    CAN_HANDLE_FILE_ACTIONS: bool = False
    """
    Whether the plugin declares it can handle file actions.

    If set to ``True``, then the :meth:`create_new_file`,
    :meth:`open_last_closed_file`, :meth:`save_file`, :meth:`save_file_as`,
    :meth:`save_copy_as`, :meth:`save_all`, :meth:`revert_file`,
    :meth:`close_file` and :meth:`close_all` methods will be called
    to handle the corresponding actions.

    If set to ``False``, the default, the corresponding actions fall back
    to calling these methods on the :guilabel:`Editor`.

    Individual actions can be disabled with the
    :meth:`spyder.plugins.application.plugin.enable_file_action` method
    in the :guilabel:`Application` plugin.
    """

    FILE_EXTENSIONS: list[str] = []
    """
    List of file extensions which the plugin can open.

    If the user opens a file with one of these extensions, then the file
    will open in this plugin using its :meth:`open_file` method.

    Examples
    --------

    For example, in the Spyder-Notebook plugin to open Jupyter notebooks:

    .. code-block:: python

        FILE_EXTENSIONS = [".ipynb"]
    """

    CAN_HANDLE_EDIT_ACTIONS: bool = False
    """
    Whether the plugin can handle editing actions.

    If set to ``True``, then the :meth:`undo`, :meth:`redo`, :meth:`cut`,
    :meth:`copy`, :meth:`paste` and :meth:`select_all` methods will be called
    to handle the corresponding actions.

    If set to ``False``, the default, the corresponding actions fall back
    to calling these methods on the :guilabel:`Editor`.

    Individual actions can be disabled with the
    :meth:`spyder.plugins.application.plugin.enable_edit_action` method
    in the :guilabel:`Application` plugin.
    """

    CAN_HANDLE_SEARCH_ACTIONS = False
    """
    Whether the plugin can handle search actions.

    If set to ``True``, then the :meth:`find`, :meth:`find_next`,
    :meth:`find_previous` and :meth:`replace` methods will be called
    to handle the corresponding actions.

    If set to ``False``, the default, the corresponding actions fall back
    to calling these methods on the :guilabel:`Editor`.

    Individual actions can be disabled with the
    :meth:`spyder.plugins.application.plugin.enable_search_action` method
    in the :guilabel:`Application` plugin.
    """

    # ---- API: Available signals
    # -------------------------------------------------------------------------
    sig_focus_changed: Signal = Signal()
    """
    Report the focus state of this plugin has changed.
    """

    sig_toggle_view_changed: Signal = Signal(bool)
    """
    Report that visibility of a dockable plugin has changed.

    This is triggered by checking/unchecking the entry for a pane in the
    :menuselection:`Window --> Panes` menu.

    Parameters
    ----------
    visible: bool
        Whether the dockwidget has been shown (``True``) or hidden (``False``).
    """

    sig_switch_to_plugin_requested: Signal = Signal(object, bool)
    """
    Request the main window show this plugin's widget.

    Parameters
    ----------
    plugin: SpyderDockablePlugin
        The plugin object to show.
    force_focus: bool
        If ``True``, always give the plugin's widget focus when showing
        or hiding it with this method. If ``False``, the default,
        don't give it focus.
    """

    sig_update_ancestor_requested: Signal = Signal()
    """
    Notify the main window that a child widget needs its ancestor updated.
    """

    # ---- Private methods
    # -------------------------------------------------------------------------
    def __init__(
        self,
        parent: spyder.app.mainwindow.MainWindow,
        configuration: spyder.config.manager.ConfigurationManager,
    ):
        """
        Initialize the plugin object (called automatically by Spyder).

        .. important::

            Plugins are initialized automatically by Spyder, so they shouldn't
            call or override this method directly.

        Parameters
        ----------
        parent : spyder.app.mainwindow.MainWindow
            Parent Spyder window of this plugin.
        configuration : spyder.config.manager.ConfigurationManager
            The Spyder configuration manager object for the plugin to access.

        Raises
        ------
        SpyderAPIError
            If :attr:`WIDGET_CLASS` is not set to a subclass of
            :class:`spyder.api.widgets.main_widget.PluginMainWidget`.

        Returns
        -------
        None
        """
        if not issubclass(self.WIDGET_CLASS, PluginMainWidget):
            raise SpyderAPIError(
                "A SpyderDockablePlugin must define a valid WIDGET_CLASS "
                "attribute!"
            )

        self.CONTAINER_CLASS = self.WIDGET_CLASS
        super().__init__(parent, configuration=configuration)

        self._switch_to_shortcut = None
        """Shortcut to switch to the plugin, defined on the main window."""

        # Widget setup
        # --------------------------------------------------------------------
        self._widget = self._container
        widget = self._widget
        if widget is None:
            raise SpyderAPIError(
                "A dockable plugin must define a WIDGET_CLASS!"
            )

        if not isinstance(widget, PluginMainWidget):
            raise SpyderAPIError(
                "The WIDGET_CLASS of a dockable plugin must be a subclass of "
                "PluginMainWidget!"
            )

        widget.DISABLE_ACTIONS_WHEN_HIDDEN = self.DISABLE_ACTIONS_WHEN_HIDDEN
        widget.RAISE_AND_FOCUS = self.RAISE_AND_FOCUS
        widget.set_icon(self.get_icon())
        widget.set_name(self.NAME)

        # Render all toolbars as a final separate step on the main window
        # in case some plugins want to extend a toolbar. Since the rendering
        # can only be done once!
        widget.render_toolbars()

        # Default Signals
        # --------------------------------------------------------------------
        widget.sig_toggle_view_changed.connect(self.sig_toggle_view_changed)
        widget.sig_update_ancestor_requested.connect(
            self.sig_update_ancestor_requested
        )

    # ---- API: Optional methods to override
    # -------------------------------------------------------------------------
    def create_new_file(self) -> None:
        """
        Create a new file inside the plugin.

        This method will be called if the user creates a new file using
        the :menuselection:`File --> New` menu item or the :guilabel:`New file`
        button in the main toolbar, and :attr:`CAN_HANDLE_FILE_ACTIONS` is
        set to ``True``.

        Returns
        -------
        None
        """
        raise NotImplementedError

    def open_file(self, filename: str) -> None:
        """
        Open a file inside the plugin.

        This method will be called if the user wants to open a file with one
        of the file name extensions listed in :attr:`FILE_EXTENSIONS`,
        so that attribute needs to be set too in order to use this method.

        Parameters
        ----------
        filename: str
            The name of the file to be opened.

        Returns
        -------
        None
        """
        raise NotImplementedError

    def get_current_filename(self) -> str | None:
        """
        Return the name of the file that is currently displayed.

        This is meant for plugins like the :guilabel:`Editor` or
        Spyder-Notebook which can display or edit files. Return ``None``
        if no file is displayed or if this plugin does not display files.

        This method is used in the :guilabel:`Open file` action to
        initialize the :guilabel:`Open file` dialog.

        Returns
        -------
        str | None
            The filename currently displayed in the plugin as a string,
            or ``None`` if no file is opened.
        """
        return None

    def current_file_is_temporary(self) -> bool:
        """
        Return whether the currently displayed file is a temporary file.

        This method should only be called if a file is displayed; that is,
        if :meth:`get_current_filename` does not return `None`.

        Returns
        -------
        bool
            ``True`` if the plugin's currently displayed file is a temporary
            one, ``False`` otherwise.
        """
        return False

    def open_last_closed_file(self) -> None:
        """
        Reopen the last closed file.

        This method will be called if the
        :menuselection:`File --> Open last closed` menu item is selected
        while the plugin has focus and :attr:`CAN_HANDLE_FILE_ACTIONS`
        is set to ``True``.

        Returns
        -------
        None
        """
        raise NotImplementedError

    def save_file(self) -> None:
        """
        Save the current file.

        This method will be called if the user saves the current file using
        the :menuselection:`File --> Save` menu item or the
        :guilabel:`Save file` button in the main toolbar, the plugin has focus,
        and :attr:`CAN_HANDLE_FILE_ACTIONS` is set to ``True``.

        Returns
        -------
        None
        """
        raise NotImplementedError

    def save_file_as(self) -> None:
        """
        Save the current file under a different name.

        This method will be called if the user saves the current file using
        the :menuselection:`File --> Save as` menu item, the plugin has focus,
        and :attr:`CAN_HANDLE_FILE_ACTIONS` is set to ``True``.

        Returns
        -------
        None
        """
        raise NotImplementedError

    def save_copy_as(self) -> None:
        """
        Save a copy of the current file under a different name.

        This method will be called if the user saves the current file using
        the :menuselection:`File --> Save copy as` menu item,
        the plugin has focus, and :attr:`CAN_HANDLE_FILE_ACTIONS` is set to
        ``True``.

        Returns
        -------
        None
        """
        raise NotImplementedError

    def save_all(self) -> None:
        """
        Save all files that are opened in the plugin.

        This method will be called if the user saves all open file using
        the :menuselection:`File --> Save all` menu item or the
        :guilabel:`Save all` button in the main toolbar, the plugin has focus,
        and :attr:`CAN_HANDLE_FILE_ACTIONS` is set to ``True``.

        Returns
        -------
        None
        """
        raise NotImplementedError

    def revert_file(self) -> None:
        """
        Revert the current file to the version stored on disk.

        This method will be called if the :menuselection:`File --> Revert`
        menu item is selected, the plugin has focus, and
        :attr:`CAN_HANDLE_FILE_ACTIONS` is set to ``True``.

        Returns
        -------
        None
        """
        raise NotImplementedError

    def close_file(self) -> None:
        """
        Close the current file.

        This method will be called if the :menuselection:`File --> Close`
        menu item is selected, the plugin has focus, and
        :attr:`CAN_HANDLE_FILE_ACTIONS` is set to ``True``.

        Returns
        -------
        None
        """
        raise NotImplementedError

    def close_all(self) -> None:
        """
        Close all opened files.

        This method will be called if the :menuselection:`File --> Close all`
        menu item is selected, the plugin has focus, and
        :attr:`CAN_HANDLE_FILE_ACTIONS` is set to ``True``.

        Returns
        -------
        None
        """
        raise NotImplementedError

    def undo(self) -> None:
        """
        Undo the most recent change.

        This method will be called if the :menuselection:`Edit --> Undo`
        menu item is selected, the plugin has focus, and
        :attr:`CAN_HANDLE_EDIT_ACTIONS` is set to ``True``.

        Returns
        -------
        None
        """
        raise NotImplementedError

    def redo(self) -> None:
        """
        Redo the most recently undone change.

        This method will be called if the :menuselection:`Edit --> Redo`
        menu item is selected, the plugin has focus, and
        :attr:`CAN_HANDLE_EDIT_ACTIONS` is set to ``True``.

        Returns
        -------
        None
        """
        raise NotImplementedError

    def cut(self) -> None:
        """
        Copy and remove the current selection.

        This method will be called if the :menuselection:`Edit --> Cut`
        menu item is selected, the plugin has focus, and
        :attr:`CAN_HANDLE_EDIT_ACTIONS` is set to ``True``.

        Returns
        -------
        None
        """
        raise NotImplementedError

    def copy(self) -> None:
        """
        Copy the current selection.

        This method will be called if the :menuselection:`Edit --> Copy`
        menu item is selected, the plugin has focus, and
        :attr:`CAN_HANDLE_EDIT_ACTIONS` is set to ``True``.

        Returns
        -------
        None
        """
        raise NotImplementedError

    def paste(self) -> None:
        """
        Paste the current clipboard contents.

        This method will be called if the :menuselection:`Edit --> Paste`
        menu item is selected, the plugin has focus, and
        :attr:`CAN_HANDLE_EDIT_ACTIONS` is set to ``True``.

        Returns
        -------
        None
        """
        raise NotImplementedError

    def select_all(self) -> None:
        """
        Select all content in the plugin.

        This method will be called if the :menuselection:`Edit --> Select all`
        menu item is selected, the plugin has focus, and
        :attr:`CAN_HANDLE_EDIT_ACTIONS` is set to ``True``.

        Returns
        -------
        None
        """
        raise NotImplementedError

    def find(self) -> None:
        """
        Search for text in the plugin.

        This method will be called if the :menuselection:`Search --> Find text`
        menu item is selected, the plugin has focus, and
        :attr:`CAN_HANDLE_SEARCH_ACTIONS` is set to ``True``.

        Returns
        -------
        None
        """
        raise NotImplementedError

    def find_next(self) -> None:
        """
        Move to the next occurrence of found text in the plugin.

        This method will be called if the :menuselection:`Search --> Find next`
        menu item is selected, the plugin has focus, and
        :attr:`CAN_HANDLE_SEARCH_ACTIONS` is set to ``True``.

        Returns
        -------
        None
        """
        raise NotImplementedError

    def find_previous(self) -> None:
        """
        Move to the previous occurrence of found text in the plugin.

        This method will be called if the
        :menuselection:`Search --> Find previous` menu item is selected,
        the plugin has focus, and :attr:`CAN_HANDLE_SEARCH_ACTIONS` is set
        to ``True``.

        Returns
        -------
        None
        """
        raise NotImplementedError

    def replace(self) -> None:
        """
        Replace occurrence of text in the plugin.

        This method will be called if the
        :menuselection:`Search --> Replace text` menu item is selected,
        the plugin has focus, and :attr:`CAN_HANDLE_SEARCH_ACTIONS` is set
        to ``True``.

        Returns
        -------
        None
        """
        raise NotImplementedError

    # ---- API: available methods
    # -------------------------------------------------------------------------
    def before_long_process(self, message: str) -> None:
        """
        Perform actions required before starting a long-running process.

        Shows a message in the main window's status bar,
        changes the mouse pointer to a wait cursor, and starts a spinner.

        Parameters
        ----------
        message: str
            Message to show in the status bar when the long process starts.

        Returns
        -------
        None
        """
        self.get_widget().start_spinner()
        super().before_long_process(message)

    def after_long_process(self, message: str = "") -> None:
        """
        Perform actions required after starting a long-running process.

        Clears the message in the Spyder main window's status bar,
        restores the mouse pointer to the OS default, and stops spinner.

        Parameters
        ----------
        message: str, optional
            Message to show in the status bar when the long process finishes.
            An empty message by default, clearing the previous message.

        Returns
        -------
        None
        """
        super().after_long_process(message)
        self.get_widget().stop_spinner()

    def get_widget(self) -> PluginMainWidget:
        """
        Return the plugin's main widget.

        Returns
        -------
        spyder.api.widgets.main_widget.PluginMainWidget
            The plugin's main widget, an instance of :attr:`WIDGET_CLASS`.
        """
        if self._widget is None:
            raise SpyderAPIError("Dockable plugin must have a WIDGET_CLASS!")

        return self._widget

    def update_title(self) -> None:
        """
        Update the plugin widget's title, i.e. its dockwidget or window title.

        Returns
        -------
        None
        """
        self.get_widget().update_title()

    def update_margins(self, margin: int | None) -> None:
        """
        Update the margins of the main widget inside this dockable plugin.

        Parameters
        ----------
        margin: int | None
            The margins to use for the main widget, or ``None`` for the
            default margins.

        Returns
        -------
        None
        """
        self.get_widget().update_margins(margin)

    @Slot()
    def switch_to_plugin(self, force_focus: bool = False) -> None:
        """
        Switch to this plugin and define if focus should be given or not.

        Parameters
        ----------
        force_focus : bool | None, optional
            If ``True``, always give the plugin's widget focus when showing
            or hiding it with this method. If ``False``, the default,
            don't give it focus.

        Returns
        -------
        None
        """
        if self.get_widget().windowwidget is None:
            self.sig_switch_to_plugin_requested.emit(self, force_focus)

    def set_ancestor(self, ancestor_widget: QWidget) -> None:
        """
        Update the ancestor/parent of child widgets when undocking.

        Parameters
        ----------
        ancestor_widget: QWidget
            The window widget to set as a parent of this one.

        Returns
        -------
        None
        """
        self.get_widget().set_ancestor(ancestor_widget)

    # ---- Convenience methods from the widget exposed on the plugin
    # -------------------------------------------------------------------------
    @property
    def dockwidget(self) -> spyder.widgets.dock.SpyderDockWidget:
        """The dockable widget (pane) for this plugin."""
        return self.get_widget().dockwidget

    @property
    def options_menu(self) -> PluginMainWidgetOptionsMenu:
        """The options ("hamburger") menu widget for this plugin."""
        return self.get_widget().get_options_menu()

    @property
    def toggle_view_action(self) -> spyder.utils.qthelpers.SpyderAction:
        """The :class:`QAction` for showing/hiding this plugin's pane."""
        return self.get_widget().toggle_view_action

    def create_dockwidget(
        self,
        mainwindow: spyder.app.mainwindow.MainWindow,
    ) -> spyder.widgets.dock.SpyderDockWidget:
        """
        Create a new dock widget for this plugin.

        Parameters
        ----------
        mainwindow : spyder.app.mainwindow.MainWindow
            The main window to set as the dockwidget's parent.

        Returns
        -------
        spyder.widgets.dock.SpyderDockWidget
            The new dock widget created for the plugin.
        """
        return self.get_widget().create_dockwidget(mainwindow)

    def create_window(self) -> None:
        """
        Create an undocked window for the plugin.

        Returns
        -------
        None
        """
        self.get_widget().create_window()

    def close_window(self, save_undocked: bool = False) -> None:
        """
        Close the plugin's undocked window, optionally saving its state.
        
        This is a convenience wrapper to close an undocked plugin.

        Parameters
        ----------
        save_undocked : bool, optional
            ``True`` if the window state (size and position) should be saved.
            If ``False``, the default, don't persist the window state.

        Returns
        -------
        None
        """
        self.get_widget()._close_window(save_undocked=save_undocked)

    def change_visibility(
        self, state: bool, force_focus: bool = False
    ) -> None:
        """
        Raise the plugin's dockwidget to the foreground, and/or grab its focus.

        Parameters
        ----------
        state : bool
            Whether the plugin's widget is being raised to the foreground
            (``True``) or set as not in the foreground (``False``).
            The latter does not actually send it to the background, but
            does configure it for not being actively shown (e.g. it disables
            its empty pane graphics).
        force_focus : bool | None, optional
            If ``True``, always give the plugin's widget keyboard focus when
            raising or un-raising it with this method. If ``None``, only give
            it focus when showing, not hiding (setting ``state`` to ``True``),
            and only if :attr:`RAISE_AND_FOCUS` is ``True``. If ``False``,
            the default, don't give it focus regardless.

        Returns
        -------
        None
        """
        self.get_widget().change_visibility(state, force_focus)

    def toggle_view(self, value: bool) -> None:
        """
        Show or hide the plugin's dockwidget in the Spyder interface.

        Used to show or hide it from the from the
        :menuselection:`Window --> Panes` menu.

        Parameters
        ----------
        value : bool
            Whether to show (``True``) or hide (``False``) the dockwidget.

        Returns
        -------
        None
        """
        self.get_widget().toggle_view(value)



# -----------------------------------------------------------------------------
# Copyright (c) 2013-2016 Colin Duquesnoy and others (see pyqode/AUTHORS.rst)
# Copyright (c) 2016- Spyder Project Contributors (see AUTHORS.txt)
#
# Released under the terms of the MIT License
# (see NOTICE.txt in the Spyder root directory for details)
# -----------------------------------------------------------------------------

"""
Mixin classes for the Spyder widget API.
"""

from __future__ import annotations

# Standard library imports
from collections import OrderedDict
from typing import Any, TYPE_CHECKING

# Third party imports
from qtpy.QtCore import QPoint, Qt
from qtpy.QtGui import QImage, QPainter, QPixmap
from qtpy.QtSvg import QSvgRenderer
from qtpy.QtWidgets import QApplication, QSizePolicy, QWidget

# Local imports
from spyder.api.config.mixins import (
    SpyderConfigurationAccessor,
    SpyderConfigurationObserver
)
from spyder.api.exceptions import SpyderAPIError
from spyder.api.shortcuts import SpyderShortcutsMixin
from spyder.api.widgets.menus import SpyderMenu
from spyder.api.widgets.toolbars import SpyderToolbar
from spyder.config.manager import CONF
from spyder.utils.icon_manager import ima
from spyder.utils.image_path_manager import get_image_path
from spyder.utils.qthelpers import create_action, create_toolbutton
from spyder.utils.registries import (
    ACTION_REGISTRY,
    MENU_REGISTRY,
    TOOLBAR_REGISTRY,
    TOOLBUTTON_REGISTRY,
)
from spyder.utils.stylesheet import PANES_TOOLBAR_STYLESHEET

if TYPE_CHECKING:
    from collections.abc import Callable

    from qtpy.QtCore import QObject
    from qtpy.QtGui import QIcon
    from qtpy.QtWidgets import QAction, QMainWindow, QToolButton

    import spyder.utils.qthelpers  # For SpyderAction
    import spyder.config.types  # For ConfigurationKey


class SpyderToolButtonMixin:
    """
    Provide methods to create, add and get toolbar buttons.
    """

    def create_toolbutton(
        self,
        name: str,
        text: str | None = None,
        icon: QIcon | str | None = None,
        tip: str | None = None,
        toggled: Callable[[Any], None] | bool | None = None,
        triggered: Callable[[], None] | None = None,
        autoraise: bool = True,
        text_beside_icon: bool = False,
        section: str | None = None,
        option: spyder.config.types.ConfigurationKey | None = None,
        register: bool = True,
    ) -> QToolButton:
        """
        Create a new Spyder toolbar button.

        Parameters
        ----------
        name: str
            Unique identifier for the toolbutton.
        text: str | None, optional
           Localized text for the toolbutton, displayed in the interface.
           If ``None``, the default, no text is shown.
        icon: QIcon | str | None, optional
            Icon name or object to use in the toolbutton.
            If ``None``, the default, no icon is shown.
        tip: str | None, optional
            Tooltip text for the toolbutton.
            If ``None`` (the default), no tooltip is shown.
        toggled: Callable[[Any], None] | bool | None, optional
            If ``None`` (default) then the button doesn't act like a checkbox.
            If ``True``, then the button modifies the configuration ``option``
            on the ``section`` specified (or the default
            :attr:`~spyder.api.plugins.SpyderPluginV2.CONF_SECTION` if
            ``section`` is not set). Otherwise, must be a callable,
            called when toggling this button. One of ``toggled`` and
            ``triggered`` must not be ``None``.
        triggered: Callable[[], None] | None, optional
            If a callable, will be called when triggering this button.
            Otherwise, if ``None`` (the default), this will not be a
            triggered button and ``toggled`` must be non-``None`` instead.
        autoraise : bool, optional
            If ``True`` (the default), the button will only draw a 3D frame
            when hovered over. If ``False``, will draw the button frame
            all the time.
        text_beside_icon : bool, optional
            If ``True``, the button text will be displayed beside the icon.
            If ``False`` (the default), will only show the icon.
        section: str | None, optional
            Name of the configuration section whose ``option`` is going to be
            modified. If ``None`` (the default) and ``option`` is not ``None``,
            then it defaults to the class'
            :attr:`~spyder.api.plugins.SpyderPluginV2.CONF_SECTION` attribute.
        option: spyder.config.types.ConfigurationKey | None, optional
            Name of the configuration option whose value is reflected and
            affected by the button. If ``None`` (the default), no option
            is associated with this button, e.g. for buttons that are
            ``triggered`` rather than ``toggled``.
        register_action: bool, optional
            If ``True`` (default) the action will be registered and searchable.
            If ``False``, the action will be created but not registered.

        Returns
        -------
        QToolButton
            The toolbar button object that was created.
        """
        if toggled and not callable(toggled):
            toggled = lambda value: None

        if toggled is not None:
            if section is None and option is not None:
                section = self.CONF_SECTION

        toolbutton = create_toolbutton(
            self,
            text=text,
            shortcut=None,
            icon=icon,
            tip=tip,
            toggled=toggled,
            triggered=triggered,
            autoraise=autoraise,
            text_beside_icon=text_beside_icon,
            section=section,
            option=option,
            id_=name,
            plugin=self.PLUGIN_NAME,
            context_name=self.CONTEXT_NAME,
            register_toolbutton=register,
        )
        toolbutton.name = name

        if toggled:
            if section is not None and option is not None:
                value = CONF.get(section, option)
                toolbutton.setChecked(value)

        return toolbutton

    def get_toolbutton(
        self,
        name: str,
        context: str | None = None,
        plugin: str | None = None,
    ) -> QToolButton:
        """
        Retrieve a toolbar button by name, context and plugin.

        Parameters
        ----------
        name: str
            Identifier of the toolbutton to retrieve.
        context: str | None, optional
            Context identifier under which the toolbutton was stored.
            If ``None``, the default, then the
            :attr:`~SpyderWidgetMixin.CONTEXT_NAME` attribute is used instead.
        plugin: str | None, optional
            Identifier of the plugin in which the toolbutton was defined.
            If ``None``, the default, then the
            :attr:`~SpyderWidgetMixin.PLUGIN_NAME` attribute is used instead.

        Returns
        -------
        QToolButton
            The corresponding toolbutton widget stored under the given
            ``name``, ``context`` and ``plugin``.

        Raises
        ------
        KeyError
            If the combination of ``name``, ``context`` and ``plugin`` keys
            does not exist in the toolbutton registry.
        """
        plugin = self.PLUGIN_NAME if plugin is None else plugin
        context = self.CONTEXT_NAME if context is None else context
        return TOOLBUTTON_REGISTRY.get_reference(name, plugin, context)

    def get_toolbuttons(
        self, context: str | None = None, plugin: str | None = None
    ) -> dict[str, QToolButton]:
        """
        Return all toolbar buttons defined by a context on a given plugin.

        Parameters
        ----------
        context: str | None, optional
            Context identifier under which the toolbuttons were stored.
            If ``None``, the default, then the
            :attr:`~SpyderWidgetMixin.CONTEXT_NAME` attribute is used instead.
        plugin: str | None, optional
            Identifier of the plugin in which the toolbuttons were defined.
            If ``None``, the default, then the
            :attr:`~SpyderWidgetMixin.PLUGIN_NAME` attribute is used instead.

        Returns
        -------
        dict[str, QToolButton]
            A dictionary that maps identifier name keys to their corresponding
            toolbutton objects.
        """
        plugin = self.PLUGIN_NAME if plugin is None else plugin
        context = self.CONTEXT_NAME if context is None else context
        return TOOLBUTTON_REGISTRY.get_references(plugin, context)


class SpyderToolbarMixin:
    """
    Provide methods to create, add and get toolbars.
    """

    def add_item_to_toolbar(
        self,
        action_or_widget: spyder.utils.qthelpers.SpyderAction | QWidget,
        toolbar: SpyderToolbar,
        section: str | None = None,
        before: str | None = None,
        before_section: str | None = None,
    ) -> None:
        """
        Add the given action or widget to this toolbar.

        Parameters
        ----------
        action_or_widget : spyder.utils.qthelpers.SpyderAction | QWidget
            The action or widget to add to the toolbar.
        toolbar : SpyderToolbar
            The toolbar object to add ``action_or_widget`` to.
        section: str | None, optional
            The section id in which to insert the ``action_or_widget``,
            or ``None`` (default) for no section.
        before: str | None, optional
            Make the ``action_or_widget`` appear before the action with the
            identifier ``before``. If ``None`` (default), add it to the end.
            If ``before`` is not ``None``, ``before_section`` will be ignored.
        before_section : str | None, optional
            Make the ``section`` appear prior to ``before_section``.
            If ``None`` (the default), add the section to the end.
            If you provide a ``before`` action, the new action will be placed
            before this one, so the section option will be ignored, since the
            action will now be placed in the same section as the ``before``
            action.

        Returns
        -------
        None
        """
        toolbar.add_item(
            action_or_widget,
            section=section,
            before=before,
            before_section=before_section,
        )

    def create_stretcher(self, id_: str | None = None) -> QWidget:
        """
        Create a stretcher widget to be used in a Spyder toolbar.

        Parameters
        ----------
        id_ : str | None, optional
            The identifier for the stretcher widget. If ``None`` (the default),
            no id will be set.

        Returns
        -------
        QWidget
            The created stretcher widget.
        """
        stretcher = QWidget(self)
        stretcher.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Expanding)

        if id_ is not None:
            stretcher.ID = id_
        return stretcher

    def create_toolbar(
        self, name: str, register: bool = True
    ) -> SpyderToolbar:
        """
        Create a Spyder toolbar.

        Parameters
        ----------
        name: str
            Unique string identifier name of the toolbar to create.
        register: bool, optional
            If ``True`` (default), register the toolbar in the global registry.
            If ``False``, don't register it.

        Returns
        -------
        SpyderToolbar
            The created toolbar object.
        """
        toolbar = SpyderToolbar(self, name)
        toolbar.setStyleSheet(str(PANES_TOOLBAR_STYLESHEET))
        if register:
            TOOLBAR_REGISTRY.register_reference(
                toolbar, name, self.PLUGIN_NAME, self.CONTEXT_NAME
            )
        return toolbar

    def get_toolbar(
        self,
        name: str,
        context: str | None = None,
        plugin: str | None = None,
    ) -> SpyderToolbar:
        """
        Return toolbar by name, context and plugin.

        Parameters
        ----------
        name: str
            Identifier of the toolbar to retrieve.
        context: str | None, optional
            Context identifier under which the toolbar was stored.
            If ``None``, the default, then the
            :attr:`~SpyderWidgetMixin.CONTEXT_NAME` attribute is used instead.
        plugin: str | None, optional
            Identifier of the plugin in which the toolbar was defined.
            If ``None``, the default, then the
            :attr:`~SpyderWidgetMixin.PLUGIN_NAME` attribute is used instead.

        Returns
        -------
        SpyderToolbar
            The corresponding toolbar widget stored under the given
            ``name``, ``context`` and ``plugin``.

        Raises
        ------
        KeyError
            If the combination of ``name``, ``context`` and ``plugin`` keys
            does not exist in the toolbar registry.
        """
        plugin = self.PLUGIN_NAME if plugin is None else plugin
        context = self.CONTEXT_NAME if context is None else context
        return TOOLBAR_REGISTRY.get_reference(name, plugin, context)

    def get_toolbars(
        self, context: str | None = None, plugin: str | None = None
    ) -> dict[str, SpyderToolbar]:
        """
        Return all toolbars defined by a context on a given plugin.

        Parameters
        ----------
        context: str | None, optional
            Context identifier under which the toolbars were stored.
            If ``None``, the default, then the
            :attr:`~SpyderWidgetMixin.CONTEXT_NAME` attribute is used instead.
        plugin: str | None, optional
            Identifier of the plugin in which the toolbars were defined.
            If ``None``, the default, then the
            :attr:`~SpyderWidgetMixin.PLUGIN_NAME` attribute is used instead.

        Returns
        -------
        dict[str, SpyderToolbar]
            A dictionary that maps identifier name keys to their corresponding
            toolbar objects.
        """
        plugin = self.PLUGIN_NAME if plugin is None else plugin
        context = self.CONTEXT_NAME if context is None else context
        return TOOLBAR_REGISTRY.get_references(plugin, context)


class SpyderMenuMixin:
    """
    Provide methods to create, add and get menus.

    This mixin uses a custom menu object that allows for the creation of
    sections in a simple way.
    """

    def add_item_to_menu(
        self,
        action_or_menu: spyder.utils.qthelpers.SpyderAction | SpyderMenu,
        menu: SpyderMenu,
        section: str | None = None,
        before: str | None = None,
        before_section: str | None = None,
    ) -> None:
        """
        Add the given action or widget to the menu.

        Parameters
        ----------
        action_or_menu : spyder.utils.qthelpers.SpyderAction | SpyderMenu
            The action or submenu to add to the menu.
        menu : SpyderMenu
            The menu object to add ``action_or_menu`` to.
        section: str | None, optional
            The section id in which to insert the ``action_or_menu``,
            or ``None`` (default) for no section.
        before: str | None, optional
            Make the ``action_or_menu`` appear before the action with the
            identifier ``before``. If ``None`` (default), add it to the end.
            If ``before`` is not ``None``, ``before_section`` will be ignored.
        before_section : str | None, optional
            Make the ``section`` appear prior to ``before_section``.
            If ``None`` (the default), add the section to the end.

        Returns
        -------
        None

        Raises
        ------
        SpyderAPIError
            If ``menu`` is not an instance of
            :class:`~spyder.api.widgets.menus.SpyderMenu`.
        """
        if not isinstance(menu, SpyderMenu):
            raise SpyderAPIError("Menu must be an instance of SpyderMenu!")

        menu.add_action(
            action_or_menu,
            section=section,
            before=before,
            before_section=before_section,
        )

    def _create_menu(
        self,
        menu_id: str,
        parent: QWidget | None = None,
        title: str | None = None,
        icon: QIcon | None = None,
        reposition: bool = True,
        register: bool = True,
        min_width: int | None = None,
        MenuClass: type[SpyderMenu] = SpyderMenu,
    ) -> SpyderMenu:
        """
        Create a SpyderMenu or a subclass of it.

        Notes
        -----
        * This method must only be used directly to generate a menu that is a
          subclass of SpyderMenu.
        * Refer to the documentation for `SpyderMenu` to learn about its args.
        """
        if register:
            menus = getattr(self, "_menus", None)
            if menus is None:
                self._menus = OrderedDict()

            if menu_id in self._menus:
                raise SpyderAPIError(
                    'Menu name "{}" already in use!'.format(menu_id)
                )

        menu = MenuClass(
            parent=self if parent is None else parent,
            menu_id=menu_id,
            title=title,
            min_width=min_width,
            reposition=reposition,
        )

        if icon is not None:
            menu.menuAction().setIconVisibleInMenu(True)
            menu.setIcon(icon)

        if register:
            MENU_REGISTRY.register_reference(
                menu, menu_id, self.PLUGIN_NAME, self.CONTEXT_NAME
            )
            self._menus[menu_id] = menu

        return menu

    def create_menu(
        self,
        menu_id: str,
        title: str | None = None,
        icon: QIcon | None = None,
        reposition: bool = True,
        register: bool = True,
    ) -> SpyderMenu:
        """
        Create a menu for Spyder.

        Parameters
        ----------
        menu_id: str
            Unique string identifier name for the menu to create.
        title: str | None, optional
            Localized title for the menu.
        icon: QIcon | None, optional
            Icon object to use for the menu.
        reposition: bool, optional
            If ``True`` (the default), vertically reposition the menu per
            its padding. If ``False``, don't reposition.
        register: bool, optional
            If ``True`` (default), register the menu in the global registry.
            If ``False``, don't register it.

        Returns
        -------
        SpyderMenu
            The created menu object.
        """
        return self._create_menu(
            menu_id=menu_id,
            title=title,
            icon=icon,
            reposition=reposition,
            register=register,
        )

    def get_menu(
        self,
        name: str,
        context: str | None = None,
        plugin: str | None = None,
    ) -> SpyderMenu:
        """
        Return a menu by name, context and plugin.

        Parameters
        ----------
        name: str
            Identifier of the menu to retrieve.
        context: str | None, optional
            Context identifier under which the menu was stored.
            If ``None``, the default, then the
            :attr:`~SpyderWidgetMixin.CONTEXT_NAME` attribute is used instead.
        plugin: str | None, optional
            Identifier of the plugin in which the menu was defined.
            If ``None``, the default, then the
            :attr:`~SpyderWidgetMixin.PLUGIN_NAME` attribute is used instead.

        Returns
        -------
        SpyderMenu
            The corresponding menu widget stored under the given
            ``name``, ``context`` and ``plugin``.

        Raises
        ------
        KeyError
            If the combination of ``name``, ``context`` and ``plugin`` keys
            does not exist in the menu registry.
        """
        plugin = self.PLUGIN_NAME if plugin is None else plugin
        context = self.CONTEXT_NAME if context is None else context
        return MENU_REGISTRY.get_reference(name, plugin, context)

    def get_menus(
        self, context: str | None = None, plugin: str | None = None
    ) -> dict[str, SpyderMenu]:
        """
        Return all menus defined by a context on a given plugin.

        Parameters
        ----------
        context: str | None, optional
            Context identifier under which the menus were stored.
            If ``None``, the default, then the
            :attr:`~SpyderWidgetMixin.CONTEXT_NAME` attribute is used instead.
        plugin: str | None, optional
            Identifier of the plugin in which the menus were defined.
            If ``None``, the default, then the
            :attr:`~SpyderWidgetMixin.PLUGIN_NAME` attribute is used instead.

        Returns
        -------
        dict[str, SpyderMenu]
            A dictionary that maps identifier name keys to their corresponding
            menu objects.
        """
        plugin = self.PLUGIN_NAME if plugin is None else plugin
        context = self.CONTEXT_NAME if context is None else context
        return MENU_REGISTRY.get_references(plugin, context)


class SpyderActionMixin:
    """
    Provide methods to create, add and get actions in a unified way.

    This mixin uses a custom action object.
    """

    def _update_action_state(self, action_name: str, value: bool) -> None:
        """
        This allows to update the state of a togglable action without emitting
        signals.

        This is useful when the global application configuration changes and
        we need to propagate the current state of an action based on those
        changes
        """
        self.blockSignals(True)
        try:
            self.get_action(action_name).setChecked(value)
        except SpyderAPIError:
            pass
        self.blockSignals(False)

    # Comment: The word `context` is used for two different concepts.
    # On one side it refers to a Qt widget shortcut context and on the
    # other it refers to a section of the configuration (or the widget
    # name where it is applied).
    def create_action(
        self,
        name: str,
        text: str,
        icon: QIcon | str | None = None,
        icon_text: str = "",
        tip: str | None = None,
        toggled: Callable[[Any], None] | bool | None = None,
        triggered: Callable[[], None] | None = None,
        data: Any | None = None,
        shortcut: str | None = None,
        shortcut_context: str | None = None,
        context: Qt.ShortcutContext = Qt.WidgetWithChildrenShortcut,
        initial: bool | None = None,
        register_shortcut: bool = False,
        section: str | None = None,
        option: spyder.config.types.ConfigurationKey | None = None,
        parent: QWidget | None = None,
        register_action: bool = True,
        overwrite: bool = False,
        context_name: str | None = None,
        menurole: QAction.MenuRole | None = None,
    ) -> spyder.utils.qthelpers.SpyderAction:
        """
        Create a new Spyder-specialized :class:`QAction`.

        .. note::

            There is no need to set a shortcut when creating an action,
            unless it's a fixed (non-customizable) one. Otherwise, set the
            ``register_shortcut`` argument to ``True``. If the shortcut is
            found in the ``shortcuts`` config section of your plugin,
            then it'll be assigned; if not, it'll be left blank for the
            user to define it in Spyder :guilabel:`Preferences`.

        Parameters
        ----------
        name: str
            Unique identifier for the action.
        text: str
           Localized text for the action, displayed in the interface.
        icon: QIcon | str | None, optional
            Icon name or object for the action when used in menus/toolbuttons.
            If ``None``, the default, no icon is shown.
        icon_text: str, optional
            Descriptive text for the action's ``icon``. If a non-empty string
            is passed, this will also disable the tooltip on this toolbutton
            if part of a toolbar. If the empty string (``""``) (the default),
            no icon text is set and the tooltip will not be disabled.
        tip: str | None, optional
            Tooltip text for the action when used in menus or toolbars.
            If ``None`` (the default), no tooltip is shown.
        toggled: Callable[[Any], None] | bool | None, optional
            If ``None`` (default) then the action doesn't act like a checkbox.
            If ``True``, then the action modifies the configuration ``option``
            on the ``section`` specified (or the default
            :attr:`~spyder.api.plugins.SpyderPluginV2.CONF_SECTION` if
            ``section`` is not set). Otherwise, must be a callable,
            called when toggling this action. One of ``toggled`` and
            ``triggered`` must not be ``None``.
        triggered: Callable[[], None] | None, optional
            If a callable, will be called when triggering this action.
            Otherwise, if ``None`` (the default), this will not be a
            triggered action and ``toggled`` must be non-``None`` instead.
        data: Any | None, optional
            Arbitrary user data to be set on the action.
            If ``None``, the default, no custom data is set.
        shortcut: str | None, optional
            A fixed (not configurable) keyboard shortcut to use for the action,
            in cases where it is not practical for the user to configure the
            shortcut. If ``None`` (the default), no fixed shortcut is set.
            For a normal configurable shortcut, you instead need to set
            ``register_shortcut`` to ``True`` and list the shortcut
            as one of the plugin's configuration options.
            See :mod:`spyder.api.shortcuts` for more details.
        shortcut_context: str | None, optional
            The context name of the fixed ``shortcut``. ``None`` (the default)
            for no context or no shortcut. Use ``"_"`` for global shortcuts (
            i.e. that can be used anywhere in the application).
        context: Qt.ShortcutContext, optional
            Set the context object for the fixed shortcut.
            By default, ``Qt.WidgetWithChildrenShortcut``.
        initial: bool | None, optional
            Set the initial state of a togglable action. This does not emit
            the toggled signal. If ``None``, the default, no initial state.
        register_shortcut: bool, optional
            If ``True``, the main window will expose the shortcut in the
            Spyder :guilabel:`Preferences`. If ``False``, the default,
            the shortcut will not be registered.
        section: str | None, optional
            Name of the configuration section whose ``option`` is going to be
            modified. If ``None`` (the default) and ``option`` is not ``None``,
            then it defaults to the class'
            :attr:`~spyder.api.plugins.SpyderPluginV2.CONF_SECTION` attribute.
        option: spyder.config.types.ConfigurationKey | None, optional
            Name of the configuration option whose value is reflected and
            affected by the action. If ``None`` (the default), no option
            is associated with this action, e.g. for actions that are
            ``triggered`` rather than ``toggled``.
        parent: QWidget | None, optional
            The parent of this widget. If ``None``, the default, uses
            this instance object itself as the parent.
        register_action: bool, optional
            If ``True`` (default) the action will be registered and searchable.
            If ``False``, the action will be created but not registered.
        overwrite: bool, optional
            If ``False`` (the default) and this action overwrites another
            with the same ``name`` and ``context_name``, raise a warning.
            Set to ``True`` to disable this warning if intentionally
            overwriting another action.
        context_name: str | None, optional
            Name of the context that holds the action when registered.
            The combination of ``name`` and ``context_name`` must be unique,
            and registering an action with the same ``name`` & `context_name``
            will raise a warning by default unless ``overwrite`` is ``True``.
        menurole: QAction.MenuRole | None, optional
            Menu role for the action. Only has an effect on macOS.

        Returns
        -------
        spyder.utils.qthelpers.SpyderAction
            The Spyder action object that was created
            (a specialized :class:`QAction`).

        Raises
        ------
        SpyderAPIError
            If both ``triggered`` and ``toggled`` are ``None``, as at least one
            must be provided; or if both ``initial`` is ``True`` and
            ``triggered`` is not ``None``, as initial values can only apply to
            ``toggled`` actions.
        """
        if triggered is None and toggled is None:
            raise SpyderAPIError(
                "Action must provide the toggled or triggered parameters!"
            )

        if parent is None:
            parent = self

        if toggled and not callable(toggled):
            toggled = lambda value: None

        if toggled is not None:
            if section is None and option is not None:
                section = self.CONF_SECTION

        action = create_action(
            parent,
            text=text,
            icon=icon,
            tip=tip,
            toggled=toggled,
            triggered=triggered,
            data=data,
            context=context,
            section=section,
            shortcut=shortcut,
            option=option,
            id_=name,
            plugin=self.PLUGIN_NAME,
            context_name=(
                self.CONTEXT_NAME if context_name is None else context_name
            ),
            register_action=register_action,
            overwrite=overwrite,
            menurole=menurole,
        )
        action.name = name
        if icon_text:
            action.setIconText(icon_text)

        action.text_beside_icon = bool(icon_text)
        action.shortcut_context = shortcut_context
        action.register_shortcut = register_shortcut
        action.tip = tip

        if initial is not None:
            if toggled:
                action.setChecked(initial)
            elif triggered:
                raise SpyderAPIError(
                    "Initial values can only apply to togglable actions!"
                )
        else:
            if toggled:
                if section is not None and option is not None:
                    value = CONF.get(section, option)
                    action.setChecked(value)

        return action

    def get_action(
        self,
        name: str,
        context: str | None = None,
        plugin: str | None = None,
    ) -> spyder.utils.qthelpers.SpyderAction:
        """
        Return an action by name, context and plugin.

        Parameters
        ----------
        name: str
            Identifier of the action to retrieve.
        context: str | None, optional
            Context identifier under which the action was stored.
            If ``None``, the default, then the
            :attr:`~SpyderWidgetMixin.CONTEXT_NAME` attribute is used instead.
        plugin: str | None, optional
            Identifier of the plugin in which the action was defined.
            If ``None``, the default, then the
            :attr:`~SpyderWidgetMixin.PLUGIN_NAME` attribute is used instead.

        Returns
        -------
        spyder.utils.qthelpers.SpyderAction
            The corresponding action widget stored under the given
            ``name``, ``context`` and ``plugin``.

        Raises
        ------
        KeyError
            If the combination of ``name``, ``context`` and ``plugin`` keys
            does not exist in the action registry.
        """
        plugin = self.PLUGIN_NAME if plugin is None else plugin
        context = self.CONTEXT_NAME if context is None else context

        return ACTION_REGISTRY.get_reference(name, plugin, context)

    def get_actions(
        self, context: str | None = None, plugin: str | None = None
    ) -> dict[str, spyder.utils.qthelpers.SpyderAction]:
        """
        Return all actions defined by a context on a given plugin.

        Parameters
        ----------
        context: str | None, optional
            Context identifier under which the actions were stored.
            If ``None``, the default, then the
            :attr:`~SpyderWidgetMixin.CONTEXT_NAME` attribute is used instead.
        plugin: str | None, optional
            Identifier of the plugin in which the actions were defined.
            If ``None``, the default, then the
            :attr:`~SpyderWidgetMixin.PLUGIN_NAME` attribute is used instead.

        Returns
        -------
        dict[str, spyder.utils.qthelpers.SpyderAction]
            A dictionary that maps identifier name keys to their corresponding
            action objects.

        Notes
        -----
        * Actions should only be created once. Creating new actions on menu
          popup is *highly* discouraged.
        * Actions can be created directly on a
          :class:`~spyder.api.widgets.main_widget.PluginMainWidget` or
          :class:`~spyder.api.widgets.main_container.PluginMainContainer`
          subclass. Child widgets can also create actions, but they need to
          subclass this class or :class:`SpyderWidgetMixin`.
        * :class:`~spyder.api.widgets.main_widget.PluginMainWidget` or
          :class:`~spyder.api.widgets.main_container.PluginMainContainer`
          will collect any actions defined in subwidgets (if defined)
          and expose them in the ``get_actions`` method at the plugin level.
        * There is no need to override this method.
        """
        plugin = self.PLUGIN_NAME if plugin is None else plugin
        context = self.CONTEXT_NAME if context is None else context
        return ACTION_REGISTRY.get_references(plugin, context)

    def update_actions(self) -> None:
        """
        Update the state of exposed actions.

        Exposed actions are actions created by the :meth:`create_action`
        method.

        Returns
        -------
        None

        Raises
        ------
        NotImplementedError
            If this method is not implemented by the plugin, as is required.
        """
        raise NotImplementedError("")


class SpyderWidgetMixin(
    SpyderActionMixin,
    SpyderMenuMixin,
    SpyderToolbarMixin,
    SpyderToolButtonMixin,
    SpyderShortcutsMixin,
):
    """
    Basic functionality for all Spyder widgets and Qt items.

    This provides a simple management of widget options, as well as Qt helpers
    for defining the actions a widget provides.
    """

    PLUGIN_NAME: str | None = None
    """
    Plugin name in the action, toolbar, toolbutton & menu registries.

    Usually the same as
    :attr:`SpyderPluginV2.NAME <spyder.api.plugins.SpyderPluginV2.NAME>`,
    but may be different from :attr:`CONTEXT_NAME`.
    """

    CONTEXT_NAME: str | None = None
    """
    The name under which to store actions, toolbars, toolbuttons and menus.

    This optional attribute defines the context name under which actions,
    toolbars, toolbuttons and menus should be registered in the
    Spyder global registry.

    If those elements belong to the global scope of the plugin, then this
    attribute should have a ``None`` value, which will use the plugin's name as
    the context scope
    """

    def __init__(
        self,
        class_parent: QObject | None = None,
        parent: QWidget | None = None,
    ) -> None:
        """
        Add methods and attributes for most Spyder widgets.

        Parameters
        ----------
        class_parent : QObject | None, optional
            The parent object of this object's class, or ``None`` (default).
            Typically the plugin object.
        parent : QWidget | None, optional
            The parent widget of this one, or ``None`` (default).

        Returns
        -------
        None
        """
        for attr in ["CONF_SECTION", "PLUGIN_NAME"]:
            if getattr(self, attr, None) is None:
                if hasattr(class_parent, attr):
                    # Inherit class_parent CONF_SECTION/PLUGIN_NAME value
                    setattr(self, attr, getattr(class_parent, attr))

        super().__init__()

    @staticmethod
    def create_icon(name: str) -> QIcon:
        """
        Retrieve an icon from Spyder's icon manager.

        Parameters
        ----------
        name: str
            The name of the icon to retrieve.

        Returns
        -------
        QIcon
            The specified icon, as a :class:`QIcon` instance.
        """
        return ima.icon(name)

    def update_style(self) -> None:
        """
        Modify the interface styling used by the plugin.

        This must be reimplemented by plugins that need to adjust their style.

        Changing from the dark to the light interface theme might
        require specific styles or stylesheets to be applied. When
        the theme is changed by the user through our :guilabel:`Preferences`,
        this method will be called for all plugins.

        Returns
        -------
        None
        """
        pass

    def update_translation(self) -> None:
        """
        Update localization of the widget.

        This method will be called recursively on all widgets on language
        change.

        Returns
        -------
        None
        """
        pass


class SpyderMainWindowMixin:
    """
    Mixin with additional functionality for Spyder's :class:`QMainWindow`\\s.
    """

    def _is_on_visible_screen(self: QMainWindow) -> bool:
        """Detect if the window is placed on a visible screen."""
        x, y = self.geometry().x(), self.geometry().y()
        qapp = QApplication.instance()
        current_screen = qapp.screenAt(QPoint(x, y))

        if current_screen is None:
            return False
        else:
            return True

    def move_to_primary_screen(self: QMainWindow) -> None:
        """
        Move the window to the primary screen if necessary.

        Returns
        -------
        None
        """
        if self._is_on_visible_screen():
            return

        qapp = QApplication.instance()
        primary_screen_geometry = qapp.primaryScreen().availableGeometry()
        x, y = primary_screen_geometry.x(), primary_screen_geometry.y()

        if self.isMaximized():
            self.showNormal()

        self.move(QPoint(x, y))

        # With this we want to maximize only the Spyder main window and not the
        # plugin ones, which usually are not maximized.
        if not hasattr(self, "is_window_widget"):
            self.showMaximized()


class SvgToScaledPixmap(SpyderConfigurationAccessor):
    """
    Mixin to transform an SVG to a QPixmap.

    The resulting :class:`QPixmap` is scaled according to the scale factor
    set by users in Spyder's :guilabel:`Preferences`.
    """

    def svg_to_scaled_pixmap(
        self,
        svg_file: str,
        rescale: float | None = None,
        in_package: bool = True,
    ) -> QPixmap:
        """
        Transform an SVG to a QPixmap.

        The resulting :class:`QPixmap` is scaled according to the scale factor
        set by users in Spyder's :guilabel:`Preferences`.

        Parameters
        ----------
        svg_file: str
            Name of or path to the SVG file to convert.
        rescale: float | None, optional
            Rescale pixmap according to a factor between 0 and 1.
            If ``None`` (default), will use the default scale factor.
        in_package: bool, optional
            If ``True`` (the default), get the SVG from the
            :file:`images` directory in the installed Spyder package.
            If ``False``, retrieve it from the specified file on disk.

        Returns
        -------
        QPixmap
            The converted rasterized image.
        """
        if in_package:
            image_path = get_image_path(svg_file)

        if self.get_conf("high_dpi_custom_scale_factor", section="main"):
            scale_factors = self.get_conf(
                "high_dpi_custom_scale_factors", section="main"
            )
            scale_factor = float(scale_factors.split(":")[0])
        else:
            scale_factor = 1

        # Get width and height
        pm = QPixmap(image_path)
        width = pm.width()
        height = pm.height()

        # Rescale but preserving aspect ratio
        if rescale is not None:
            aspect_ratio = width / height
            width = int(width * rescale)
            height = int(width / aspect_ratio)

        # Paint image using svg renderer
        image = QImage(
            int(width * scale_factor),
            int(height * scale_factor),
            QImage.Format_ARGB32_Premultiplied,
        )
        image.fill(0)
        painter = QPainter(image)
        renderer = QSvgRenderer(image_path)
        renderer.render(painter)
        painter.end()

        # This is also necessary to make the image look good for different
        # scale factors
        if scale_factor > 1.0:
            image.setDevicePixelRatio(scale_factor)

        # Create pixmap out of image
        final_pm = QPixmap.fromImage(image)
        final_pm = final_pm.copy(
            0, 0, int(width * scale_factor), int(height * scale_factor)
        )

        return final_pm



# -*- coding: utf-8 -*-
# -----------------------------------------------------------------------------
# Copyright (c) 2001-2015 IPython Development Team
# Copyright (c) 2015 Jupyter Development Team
# Copyright (c) 2015- Spyder Project Contributors
#
# Distributed under the terms of the Modified BSD License
# (BSD 3-clause; see NOTICE.txt in the Spyder root directory for details).
# -----------------------------------------------------------------------------

"""
A generic Emacs-style kill ring, as well as a Qt-specific version.

Adapted from qtconsole/kill_ring.py of the
`Jupyter QtConsole Project <https://github.com/jupyter/qtconsole>`_.
"""

# Third party imports
from qtpy.QtCore import QObject
from qtpy.QtGui import QTextCursor
from qtpy.QtWidgets import QTextEdit, QPlainTextEdit


# ----------------------------------------------------------------------------
# Classes
# ----------------------------------------------------------------------------
class KillRing(object):
    """ A generic Emacs-style kill ring.
    """

    def __init__(self):
        self.clear()

    def clear(self):
        """ Clears the kill ring.
        """
        self._index = -1
        self._ring = []

    def kill(self, text):
        """ Adds some killed text to the ring.
        """
        self._ring.append(text)

    def yank(self):
        """ Yank back the most recently killed text.

        Returns
        -------
        A text string or None.
        """
        self._index = len(self._ring)
        return self.rotate()

    def rotate(self):
        """ Rotate the kill ring, then yank back the new top.

        Returns
        -------
        A text string or None.
        """
        self._index -= 1
        if self._index >= 0:
            return self._ring[self._index]
        return None


class QtKillRing(QObject):
    """ A kill ring attached to Q[Plain]TextEdit.
    """

    # -------------------------------------------------------------------------
    # QtKillRing interface
    # -------------------------------------------------------------------------

    def __init__(self, text_edit):
        """ Create a kill ring attached to the specified Qt text edit.
        """
        assert isinstance(text_edit, (QTextEdit, QPlainTextEdit))
        super().__init__()

        self._ring = KillRing()
        self._prev_yank = None
        self._skip_cursor = False
        self._text_edit = text_edit

        text_edit.cursorPositionChanged.connect(self._cursor_position_changed)

    def clear(self):
        """ Clears the kill ring.
        """
        self._ring.clear()
        self._prev_yank = None

    def kill(self, text):
        """ Adds some killed text to the ring.
        """
        self._ring.kill(text)

    def kill_cursor(self, cursor):
        """ Kills the text selected by the give cursor.
        """
        text = cursor.selectedText()
        if text:
            cursor.removeSelectedText()
            self.kill(text)

    def yank(self):
        """ Yank back the most recently killed text.
        """
        text = self._ring.yank()
        if text:
            self._skip_cursor = True
            cursor = self._text_edit.textCursor()
            cursor.insertText(text)
            self._prev_yank = text

    def rotate(self):
        """ Rotate the kill ring, then yank back the new top.
        """
        if self._prev_yank:
            text = self._ring.rotate()
            if text:
                self._skip_cursor = True
                cursor = self._text_edit.textCursor()
                cursor.movePosition(QTextCursor.Left,
                                    QTextCursor.KeepAnchor,
                                    n=len(self._prev_yank))
                cursor.insertText(text)
                self._prev_yank = text

    # -------------------------------------------------------------------------
    # Protected interface
    # -------------------------------------------------------------------------

    # ----- Signal handlers ---------------------------------------------------

    def _cursor_position_changed(self):
        if self._skip_cursor:
            self._skip_cursor = False
        else:
            self._prev_yank = None
